/*
 * Sonar Delphi Plugin
 * Copyright (C) 2019 Integrated Application Development
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.spring4d.checks;

import com.google.common.base.Splitter;
import com.google.common.collect.Iterables;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import org.sonar.check.Rule;
import org.sonar.check.RuleProperty;
import org.sonar.plugins.communitydelphi.api.ast.ForLoopVarDeclarationNode;
import org.sonar.plugins.communitydelphi.api.ast.FormalParameterListNode;
import org.sonar.plugins.communitydelphi.api.ast.FormalParameterNode.FormalParameterData;
import org.sonar.plugins.communitydelphi.api.ast.InterfaceSectionNode;
import org.sonar.plugins.communitydelphi.api.ast.NameDeclarationNode;
import org.sonar.plugins.communitydelphi.api.ast.RoutineDeclarationNode;
import org.sonar.plugins.communitydelphi.api.ast.VarDeclarationNode;
import org.sonar.plugins.communitydelphi.api.ast.VarSectionNode;
import org.sonar.plugins.communitydelphi.api.ast.VarStatementNode;
import org.sonar.plugins.communitydelphi.api.check.DelphiCheck;
import org.sonar.plugins.communitydelphi.api.check.DelphiCheckContext;
import org.sonar.plugins.communitydelphi.api.symbol.declaration.RoutineNameDeclaration;
import org.sonar.plugins.communitydelphi.api.symbol.scope.FileScope;
import org.sonar.plugins.communitydelphi.api.symbol.scope.UnitScope;
import org.sonar.plugins.communitydelphi.api.type.Type;
import org.sonar.plugins.communitydelphi.api.type.Type.ScopedType;

@Rule(key = "VariableName")
public class VariableNameCheck extends DelphiCheck {
  private static final String MESSAGE =
      "Rename this variable to match the expected naming convention.";

  @RuleProperty(
      key = "globalPrefixes",
      description =
          "Comma-delimited list of prefixes, global variables must begin with one of these.")
  public String globalPrefixes = "";

  private List<String> globalPrefixesList;

  @Override
  public void start(DelphiCheckContext context) {
    globalPrefixesList = Splitter.on(',').trimResults().splitToList(globalPrefixes);
  }

  @Override
  public DelphiCheckContext visit(VarDeclarationNode varDeclaration, DelphiCheckContext context) {
    if (isGeneratedFormVariable(varDeclaration)) {
      return context;
    }

    boolean globalVariable = varDeclaration.getScope() instanceof UnitScope;
    varDeclaration.getNameDeclarationList().getDeclarations().stream()
        .filter(node -> isViolation(node, globalVariable))
        .forEach(node -> reportIssue(context, node, MESSAGE));

    return context;
  }

  @Override
  public DelphiCheckContext visit(VarStatementNode varStatement, DelphiCheckContext context) {
    varStatement.getNameDeclarationList().getDeclarations().stream()
        .filter(node -> isViolation(node, false))
        .forEach(node -> reportIssue(context, node, MESSAGE));
    return context;
  }

  @Override
  public DelphiCheckContext visit(ForLoopVarDeclarationNode loopVar, DelphiCheckContext context) {
    NameDeclarationNode name = loopVar.getNameDeclarationNode();
    if (isViolation(name, false)) {
      reportIssue(context, name, MESSAGE);
    }
    return context;
  }

  @Override
  public DelphiCheckContext visit(
      FormalParameterListNode parameterListNode, DelphiCheckContext context) {
    var parameterNodes =
        parameterListNode.getParameters().stream()
            .map(FormalParameterData::getNode)
            .collect(Collectors.toUnmodifiableList());

    var routineNode = parameterListNode.getFirstParentOfType(RoutineDeclarationNode.class);
    var routineDeclaration = routineNode == null ? null : routineNode.getRoutineNameDeclaration();

    getParametersToCheck(routineDeclaration, parameterNodes).stream()
        .filter(parameter -> isViolation(parameter, false))
        .forEach(parameter -> reportIssue(context, parameter, MESSAGE));

    return context;
  }

  private static List<NameDeclarationNode> getParametersToCheck(
      RoutineNameDeclaration routineDeclaration, List<NameDeclarationNode> parameterNodes) {
    if (routineDeclaration == null) {
      return parameterNodes;
    }

    List<Set<String>> paramNamesFromInterfaces =
        parameterNodes.stream()
            .map(x -> new HashSet<String>())
            .collect(Collectors.toUnmodifiableList());

    findImplementedInterfaceMethodDeclarations(routineDeclaration).stream()
        .map(RoutineNameDeclaration::getParameters)
        .forEach(
            params -> {
              assert (params.size() == paramNamesFromInterfaces.size());
              for (int i = 0; i < paramNamesFromInterfaces.size(); i++) {
                paramNamesFromInterfaces.get(i).add(params.get(i).getImage());
              }
            });

    return IntStream.range(0, parameterNodes.size())
        .filter(
            index -> {
              var actualImage = parameterNodes.get(index).getImage();
              return !paramNamesFromInterfaces.get(index).contains(actualImage);
            })
        .mapToObj(parameterNodes::get)
        .collect(Collectors.toUnmodifiableList());
  }

  private boolean isViolation(NameDeclarationNode name, boolean globalVariable) {
    String image = name.getImage();
    if (globalVariable) {
      return !isPascalCaseWithPrefix(image, globalPrefixesList);
    }
    return Character.isUpperCase(image.charAt(0));
  }

  private static boolean isGeneratedFormVariable(VarDeclarationNode varDecl) {
    VarSectionNode varSection = varDecl.getVarSection();
    if (varSection.getFirstParentOfType(InterfaceSectionNode.class) == null) {
      return false;
    } else if (!(varSection.getScope() instanceof FileScope)) {
      return false;
    } else if (varSection.getDeclarations().size() != 1) {
      return false;
    } else {
      return varSection.getChildIndex() == varSection.getParent().getChildren().size() - 1
          && varDecl.getType().isDescendantOf("System.Classes.TComponent");
    }
  }

  private static Set<RoutineNameDeclaration> findImplementedInterfaceMethodDeclarations(
      RoutineNameDeclaration routine) {
    var typeDeclaration = routine.getTypeDeclaration();
    if (typeDeclaration == null) {
      return new HashSet<>();
    }

    Set<RoutineNameDeclaration> implementedMethods = new HashSet<>();

    var interfaces =
        typeDeclaration.getType().ancestorList().stream()
            .filter(Type::isInterface)
            .map(ScopedType.class::cast)
            .collect(Collectors.toUnmodifiableList());

    for (var interfaceType : interfaces) {
      interfaceType.typeScope().getRoutineDeclarations().stream()
          .filter(interfaceMethod -> hasSameSignature(routine, interfaceMethod))
          .forEach(implementedMethods::add);
    }

    return implementedMethods;
  }

  private static boolean hasSameSignature(
      RoutineNameDeclaration method, RoutineNameDeclaration overriddenMethod) {
    return method.getName().equalsIgnoreCase(overriddenMethod.getName())
        && method.hasSameParameterTypes(overriddenMethod);
  }

  private static boolean isPascalCaseWithPrefix(
      final String name, final Iterable<String> prefixes) {
    if (Iterables.isEmpty(prefixes)) {
      return isPascalCaseWithPrefix(name, "");
    }

    for (final String prefix : prefixes) {
      if (VariableNameCheck.isPascalCaseWithPrefix(name, prefix)) {
        return true;
      }
    }

    return false;
  }

  private static boolean isPascalCaseWithPrefix(final String name, final String prefix) {
    if (!name.startsWith(prefix)) {
      return false;
    }

    if (name.length() == prefix.length()) {
      return false;
    }

    char character = name.charAt(prefix.length());
    return Character.isUpperCase(character) || Character.isDigit(character);
  }
}
